/*
 * PCI Driver - Ported for Tru64 UNIX 5.1A  (October 2002)
 */

/******************************************************************
 *   /usr/sys/io/datum/pci/stfp.c
 *   Hardware Character Device Driver for Digital Unix 4.f   (June.1999)
 *   Developed at Northrop Grumman ESSS.  
 *   Programmed by  J. Kennedy and L. Goering                                                               
 *****************************************************************/
 
#include <sys/param.h> 
#include <sys/systm.h> 
#include <sys/ioctl.h> 
#include <sys/user.h> 
#include <sys/proc.h> 
#include <sys/map.h> 
#include <sys/buf.h> 
#include <sys/vm.h> 
#include <sys/file.h> 
#include <sys/uio.h> 
#include <sys/types.h>
#include <sys/errno.h>
#include <sys/conf.h>
#include <sys/kernel.h> 
#include <sys/devio.h> 
#include <hal/cpuconf.h> 
#include <sys/exec.h> 
#include <sys/sysconfig.h>
#include <machine/cpu.h> 
#include <sys/malloc.h>
#include <io/common/devdriver.h>
#include <io/dec/pci/pci.h>
#include <sys/vmmac.h>

#include "stfp_io.h" 
#include "stfp_reg.h" 


/* ************************************************************ *
 *								*
 *			W A R N I N G ! ! !			*
 *								*
 * Please check the CONTROL defintion before testing.		*
 *								*	
 * ************************************************************ */

#define 	CONTROL 	CONTROL_REG


/* ====> */

long reg_data;

#define ST_RESET 0001    /* device ready for data transfer */
#define ST_ERROR 0002    /* indicates error */

#define ST_OPEN  1       /* device open bit */
#define ST_CLOSE 0       /* device close bit */

/* <==== */


/* NO_DEV and MAJOR_INSTANCE are used as parameters in the
 * configuration stage. MAX_STFP is the maximum number of 
 * controllers that this driver supports. STFP_NAME is the 
 * name used for this driver subsystem. Be consistent! 
 */
#define NO_DEV -1
#define MAJOR_INSTANCE 1
#define MAX_STFP 16
#define STFP_NAME "stfp"

extern int nodev(), nulldev(); 

/* forward function declarations */

int stfp_probe();
int stfp_cattach();
int stfp_ctlr_unattach();

int stfp_open();
int stfp_close();
int stfp_read();
int stfp_write();
int stfp_ioctl();
int stfp_mmap();
int stfp_intr();

void stfp_preconfig(int, int, ulong, ulong);
void stfp_postconfig(int, int, ulong, ulong);

/* ===> */

void rdTimeTv(int unit, uint pTmReg, stfp_timeval_t *pTmT);
void rdTime(int unit, uint pTmReg, io_handle_t pY, stfp_time_t *pTmS);
int sendCommand(int unit);
int init (struct controller *ctlr);

unsigned int STFPRegGetL(io_handle_t regoffset);

/* <=== */


typedef struct stfp_softc {
   vm_offset_t iobase;	
   int is_open;

/* ===> */

   int                    sc_openf;      /* open flag */
   int                    sc_unit;       /* logical unit number */
   /* device_info_t          *dip;          device info */
   vm_offset_t              pDpr;        /* DPRAM base pointer */
   vm_offset_t              pRegs;
   vm_offset_t              pIn;         /* usr input area in DPRAM */
   vm_offset_t              pOut;        /* output area (to user) in DPRAM */
   vm_offset_t              pGps;        /* GPS packet area in DPRAM */
   vm_offset_t              kseg_start;
   vm_offset_t              pYear;       /* year area in DPRAM */
   struct stfp_time	IntTime;     /* Time storage structure */
   struct getdata_t         getdata;
   struct stfp_time	*IntTimePtr;
   long                     interrupt_count;
   int                      signal;      /* signal no. */
   int                      enablesignal;
   int                      waiting;     /* waiting flag */
   int                    intServiced;   /* last interrupt(s) seviced */
   int                    intMaskSig;    /* Interrupt mask for signals */
   int                    *pref;         /* process reference for signals */
   simple_lock_data_t     stfpLock;      /* Spin lock for scp access      */
   pthread_cond_t         cv;            /* cond variable (sleep/wakeup) */
   ihandler_id_t          *sc_hid;       /* interrupt block cookie */
   simple_lock_data_t     sc_intLock;    /* spin lock used for stfp_intr() */
   io_handle_t            Dprhandle;
   io_handle_t            Regshandle;
   struct proc            *proc_p;        /* pointer to parent process */
   struct controller      *sc_ctlr;       /* controller info address */

/* <=== */

} stfp_softc_t;

/* Declare an array of pointers to the structures
* used for the driver. The controller structure is
* 'created' during 'pre-configuration' and the 
* the softc structures are allocated in probe 
* ( probe is called once for every controller). */

stfp_softc_t *stfp_softc[MAX_STFP];
struct controller *stfp_info[MAX_STFP];

int stfp_num_installed=1;
int stfp_num_probed=0;
static int preconfig_status;
static int driver_cfg_state;             // dynamic or static?
static int already_configured = FALSE;      // are we already configured?

static int majnum = NO_DEV; 
static char stfp_version[] = "1.00";
static int stfp_developer_debug = 1; 
static int stfp_num_intr=0;
static char mcfgname[] = STFP_NAME; 
static char unused[300] = ""; 
static char stfp_desc[100] = ""; 
static char optiondata[400] = ""; 

/* structure used by cfgmgr in configuration */
struct driver stfpdriver = {
        stfp_probe,
        0,               /* slave */
        stfp_cattach,
        0,               /* dattach */
        0,               /* go */
        0,               /* addr_list */
        0,               /* dev_name */
        0,               /* dev_list */
        STFP_NAME,        /* ctlr_name */
        stfp_info,        /* ctlr_list */
        0,               /* xclu */
        0,               /* addr1_size */
        0,               /* addr1_atype */
        0,               /* addr2_size */
        0,               /* addr2_atype */
        stfp_ctlr_unattach,
        0                /* dev_unattach */
};

/* Struct used by kernel to map user calls to 
*  this driver's functions */
struct dsent stfp_devsw_entry = {
        stfp_open,
        stfp_close,
        nodev,		 /* strategy */
        stfp_read,
        stfp_write,
        stfp_ioctl,
        nodev,           /* d_dump */
        nodev,           /* d_psize */
        nulldev,         /* d_stop */
        nulldev,         /* d_reset */
        nulldev,         /* d_select */
        stfp_mmap,      /* d_mmap */
        0,               /* d_segmap */
        NULL,            /* d_ttys */
        DEV_FUNNEL,      /* d_funnel */
        0,               /* d_bflags */
        0                /* d_cflags */
}; 

cfg_subsys_attr_t stfp_attributes[] = { 

{"Subsystem_Description",CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY, 
                        (caddr_t)stfp_desc,2,300,0},
{"Module_Config_Name",  CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY, 
                        (caddr_t)mcfgname,2,CFG_ATTR_NAME_SZ,0},
{"majnum",              CFG_ATTR_INTTYPE, CFG_OP_QUERY,
                        (caddr_t)&majnum,0,512,0},
{"STFP_Developer_Debug", CFG_ATTR_INTTYPE, CFG_OP_QUERY |
                        CFG_OP_RECONFIGURE |
                        CFG_OP_CONFIGURE,
                        (caddr_t)&stfp_developer_debug,0,1,0}, 
{"Num_Installed",       CFG_ATTR_INTTYPE, CFG_OP_QUERY | CFG_OP_CONFIGURE,
                        (caddr_t)&stfp_num_installed,1,MAX_STFP,0},
{"Num_Probed",          CFG_ATTR_INTTYPE, CFG_OP_QUERY,
                        (caddr_t)&stfp_num_probed,0,MAX_STFP,0},
{"Module_Type",         CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,2,CFG_ATTR_NAME_SZ,0},
{"Device_Dir",          CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,0,300,0},
{"Device_Major_Req",    CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,0,300,0},
{"Device_Char_Major",   CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,0,300,0},
{"Device_Char_Files",   CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,0,300,0},
{"Device_Char_Minor",   CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)unused,0,300,0},
{"PCI_Option",          CFG_ATTR_STRTYPE, CFG_OP_CONFIGURE | CFG_OP_QUERY,
                        (caddr_t)optiondata,0,500,0}, 
/* query-only info */
{"version",             CFG_ATTR_STRTYPE, CFG_OP_QUERY,
                        (caddr_t)stfp_version,0,5,0},
{"stfp_num_intr",        CFG_ATTR_INTTYPE, CFG_OP_QUERY,
                        (caddr_t)&stfp_num_intr,0,9999999,0},
{0,0,0,0,0,0,0} 
};


/* ===> */

/*
 *========================================================================
 * Configuraion constants
 *========================================================================
 */

typedef char            int8_t;
typedef short           int16_t;
typedef int             int32_t;
typedef unsigned char   uint8_t;
typedef unsigned short  uint16_t;
typedef unsigned int    uint32_t;

/*
 *------------------------------------------------------------------------
 * STFPSLEEP is the priority level at which the driver will sleep during
 * timeouts. If you don't want a signal (such as happen when you type a ^C
 * on the keyboard) from interrupting the wait, then uncomment the
 * definition below *without* "PCATCH". However, if you don't allow
 * signals to interrupt the wait and the event never happens, you'll have
 * to reboot the system in order to kill your applications program.
 *------------------------------------------------------------------------
 */

#if    1
#define    STFPSLEEP    (PCATCH|PZERO)        /* ^C interrupts wait     */
#else     /* NOPE */
#define STFPSLEEP    (PZERO-1)                /* Nothing interrupts     */
#endif    /* NOPE */


/*
 *------------------------------------------------------------------------
 * Coordination between the bottom (interrupt level) half of the driver
 * and the upper (base level) half is coordinated by a lock and a
 * synchronization variable.
 *------------------------------------------------------------------------
 */

#define STFP_INIT_LOCK(lock_ptr)          simple_lock_init ( lock_ptr )
#define STFP_DEL_LOCK(lock_ptr)           simple_lock_terminate ( lock_ptr )
#define STFP_LOCK(lock_ptr)               simple_lock( lock_ptr )
#define STFP_UNLOCK(lock_ptr)             simple_unlock( lock_ptr )

#define STFP_WAIT_UNLOCK(sem_ptr,timeout,lock_ptr) mpsleep(sem_ptr,         \
                                                          STFPSLEEP,        \
                                                          0,               \
                                                          timeout,         \
                                                (simple_lock_t *)lock_ptr, \
                                             MS_LOCK_SIMPLE|MS_LOCK_ON_ERROR)

#define STFP_WAIT(sem_ptr, timeout)   mpsleep(sem_ptr, STFPSLEEP, 0, timeout, 0, 0)

/*----------------------------  MACROS  -------------------------------------*/


#define STFP_WRT_BUS_D8(addr, data)                            \
    {                                                          \
        mb();                                                  \
        WRITE_BUS_D8((io_handle_t)(addr),  (ulong)(data));     \
        mb();                                                  \
    }
#define STFP_WRT_BUS_D16(addr, data)                           \
    {                                                          \
        mb();                                                  \
        WRITE_BUS_D16((io_handle_t)(addr),  (ulong)(data));    \
        mb();                                                  \
    }
#define STFP_WRT_BUS_D32(addr, data)                           \
    {                                                          \
        mb();                                                  \
        WRITE_BUS_D32((io_handle_t)(addr),  (ulong)(data));    \
        mb();                                                  \
    }

#define STFP_RD_BUS_D8(addr, data)                             \
    {                                                          \
        mb();                                                  \
        *data = READ_BUS_D8((io_handle_t)(addr));              \
        mb();                                                  \
    }

#define STFP_RD_BUS_D16(addr, data)                            \
    {                                                          \
        mb();                                                  \
        *data = READ_BUS_D16((io_handle_t)(addr));             \
        mb();                                                  \
    }

#define STFP_RD_BUS_D32(addr, data)                            \
    {                                                          \
        mb();                                                  \
        *data = READ_BUS_D32((io_handle_t)(addr));             \
        mb();                                                  \
    }


#define STFP_DP_PUTB(p,val)                 \
    {                                       \
        mb();                               \
        WRITE_BUS_D8((io_handle_t)(scp->Dprhandle+p),(uchar)val); \
        mb();                               \
    }


#define STFP_REG_PUTL(p,val)                   \
    {                                          \
        mb();                                  \
        WRITE_BUS_D32((io_handle_t)(scp->Regshandle+p),(ulong)val);  \
        mb();                                  \
    }



#define STFP_RD_BUS_D32(addr, data)                            \
    {                                                          \
        mb();                                                  \
        *data = READ_BUS_D32((io_handle_t)(addr));             \
        mb();                                                  \
    }



#define STFP_REG_OREQUAL(p,val)               \
    {                                         \
     reg_data = STFPRegGetL(scp->Regshandle+p) | val;       \
     STFP_REG_PUTL(p,reg_data);               \
    }


#define STFP_REG_ANDEQUAL(p,val)                      \
    {                                         \
     reg_data = STFPRegGetL(scp->Regshandle+p) & val;       \
     STFP_REG_PUTL(p,reg_data);               \
    }

#define STFP_INTR_PENDING                        \
      {                                          \
        (STFPRegGetL(scp->Regshandle+MASK) &     \
         STFPRegGetL(scp->Regshandle+INTSTAT) &  \
                               INT_BITMASK);     \
      }


unsigned int STFPRegGetL(io_handle_t regs_handle)
{
unsigned int reg_data;
STFP_RD_BUS_D32(regs_handle, &reg_data);
return(reg_data);
}

/*
 *------------------------------------------------------------------------
 * uioDump: decode UIO structure and write to syslog
 *------------------------------------------------------------------------
 */

static  void uioDump
(
    struct stfp_softc  *scp,      /* Per-board info         */
    struct uio   *uio,      /* User-I/O structure     */
    char         *why       /* Kind of I/O being done */
)
{
    int       i;            /* Generic loop counter   */
    char   *space;          /* Address space name     */

    switch( uio->uio_segflg )
    {
        default:            space = "UNKNOWN";          break;
        case UIO_SYSSPACE:  space = "UIO_SYSSPACE";     break;
        case UIO_USERSPACE: space = "UIO_USERSPACE";    break;
    }

    printf( "%s%d:  %s UIO dump\n", scp->sc_ctlr->ctlr_name,
                                   scp->sc_ctlr->ctlr_num,
                                   why );

    printf( "%s%d:  uio_segflg = %s\n",  scp->sc_ctlr->ctlr_name,
                                        scp->sc_ctlr->ctlr_num,
                                        space );

    printf( "%s%d:  uio_fmode = 0%o\n", scp->sc_ctlr->ctlr_name,
                                       scp->sc_ctlr->ctlr_num,
                                       uio->uio_fmode );

    printf( "%s%d:  uio_offset = 0x%lX\n", scp->sc_ctlr->ctlr_name,
                                          scp->sc_ctlr->ctlr_num,
                                          (long) uio->uio_offset );

    printf( "%s%d:  uio_resid = 0x%lX\n", scp->sc_ctlr->ctlr_name,
                                         scp->sc_ctlr->ctlr_num,
                                         (long) uio->uio_resid );

    printf( "%s%d:  uio_iovcnt = 0x%X\n", scp->sc_ctlr->ctlr_name,
                                         scp->sc_ctlr->ctlr_num,
                                         uio->uio_iovcnt );

    for( i = 0; i < uio->uio_iovcnt; ++i )
    {
        struct iovec  *iov = &uio->uio_iov[i];
        printf( "%s%d:  uio_iov[%d] = (base 0x%lX, len 0x%lX)\n",
                            scp->sc_ctlr->ctlr_name,
                            scp->sc_ctlr->ctlr_num,
                            i, (long) iov->iov_base, (long) iov->iov_len );
    }

}

/*
 *------------------------------------------------------------------------
 * interfaceLock: lock a basic lock, return the cookie
 *------------------------------------------------------------------------
 */

static int interfaceLock(
    struct stfp_softc *scp,            /* Per-device info */
    register simple_lock_data_t *lock_ptr        /* lock to use     */
)
{
    spl_t    pl;    /* Priority level     */

     if (stfp_developer_debug)
    {
        printf( "%s%d: locking interface and raise priority \n",
                 scp->sc_ctlr->ctlr_name,
                 scp->sc_ctlr->ctlr_num );
    }
    pl = spldevhigh();
    STFP_LOCK( lock_ptr );
    return( pl );
}

/*
 *------------------------------------------------------------------------
 * interfaceUnlock: release lock on the interface
 *------------------------------------------------------------------------
 */

static void
interfaceUnlock(
    struct stfp_softc *scp,      /* Per-device info                 */
    register simple_lock_data_t *lock_ptr, /* lock to use                     */
    register spl_t               pri       /* priority from "interfaceLock()" */
)
{
         if (stfp_developer_debug)
        {
        printf( "%s%d: unlocking interface and return priority\n",
                         scp->sc_ctlr->ctlr_name,
                         scp->sc_ctlr->ctlr_num );
    }
    splx(pri);
    STFP_UNLOCK( lock_ptr );
}

/*
 *------------------------------------------------------------------------
 * interfaceWait: wait on synchronization variable
 *------------------------------------------------------------------------
 */

static int
interfaceWait(
    struct stfp_softc  *scp,         /* Per-device info         */
    register int                 timeout,     /* timeout                 */
    register caddr_t             sema_ptr,    /* semaphore               */
    register simple_lock_data_t *lock_ptr,    /* lock                    */
    register spl_t               pl           /* Level to grab lock      */
)
{
    register int status;

         if (stfp_developer_debug)
        {
        printf( "%s%d: unlock interface and wait on synch variable\n",
                         scp->sc_ctlr->ctlr_name,
                         scp->sc_ctlr->ctlr_num );
    }

    /* lower priority while waiting, still locked    */

    splx(pl);

    /* issues wait, unlocks, locks back after wait over */

    status = STFP_WAIT_UNLOCK(sema_ptr, timeout, lock_ptr);

    pl = spldevhigh();         /* raise priority backup                            */

         if (stfp_developer_debug)
        {
        printf( "%s%d: lock interface got the sync variable\n",
                         scp->sc_ctlr->ctlr_name,
                         scp->sc_ctlr->ctlr_num );
    }
    return( status );
}

/* <=== */


int
stfp_configure(cfg_op_t op, cfg_attr_t *indata, size_t indatalen, 
                 cfg_attr_t *outdata, size_t outdatalen)        
{
   int retval,i;
   
   if (stfp_developer_debug)
      uprintf("stfp_configure: Entry!\n");

   switch(op) {
    case CFG_OP_CONFIGURE:

      if (stfp_developer_debug)
         uprintf("- In Configure -\n");

      if (already_configured) return EINVAL;

      if(cfgmgr_get_state(mcfgname, &driver_cfg_state) != ESUCCESS){
         uprintf("stfp_conf:Error determining state of subsystem\n");
         return(EINVAL);
      }
         
      preconfig_status = ESUCCESS;      
      
      register_callback(stfp_preconfig,
                        CFG_PT_PRECONFIG,
                        CFG_ORD_NOMINAL, (long) 0L );      

      if(driver_cfg_state == SUBSYSTEM_DYNAMICALLY_CONFIGURED)  {      

         retval = configure_driver(mcfgname, DRIVER_WILDNUM,
                                   DRIVER_WILDNAME, &stfpdriver);
         if (retval != ESUCCESS) {
            if (stfp_developer_debug)
               uprintf("configure_driver failed %d\n", retval);
            return(retval);
         }

      }

      register_callback(stfp_postconfig,
                        CFG_PT_POSTCONFIG,
                        CFG_ORD_NOMINAL, (long) 0L );
      break;
    case CFG_OP_QUERY:

      if (stfp_developer_debug)
         uprintf("- In Query -\n");

      break;
    case CFG_OP_UNCONFIGURE:
      
      if (stfp_developer_debug)
         uprintf("- In UnConfigure -\n");

      if(driver_cfg_state == SUBSYSTEM_DYNAMICALLY_CONFIGURED)  {      
         
         for(i=0;i<stfp_num_probed;i++) {
            if (stfp_softc[i] && stfp_softc[i]->is_open) {
               return EBUSY;
            }
         }
         
         retval = devsw_del(mcfgname,MAJOR_INSTANCE);
         if (retval == NO_DEV) {
            if (stfp_developer_debug)
               uprintf("stfp_configure: call to devsw_del failed\n");
            return ESRCH;
         }
         
         retval = unconfigure_driver(DRIVER_WILDNAME, DRIVER_WILDNUM,
                                     &stfpdriver, DRIVER_WILDNAME,
                                     DRIVER_WILDNUM);
         if (retval != 0){
            if (stfp_developer_debug)
               uprintf("stfp_configure: unconfigure failed\n");
            return ESRCH;
         }

         already_configured = FALSE;
         
         if (stfp_developer_debug)
            uprintf("stfp_configure: dynamic module unconfigured\n");

      }
      else {
         /* static configuration */
         uprintf("stfp_configure: Can't unconfigure a static mod\n");
         return ESUCCESS;
      }
      break;
    case CFG_OP_RECONFIGURE:

      if (stfp_developer_debug)
         uprintf("- In ReConfigure -\n");

      break;
    default:
      return ENOTSUP;
   }
      
   return ESUCCESS;
}

/*---------------------------------------------------------
* pre-configuration
* 1 - Create controller structs 
* 2 - Create device structs if necessary
*--------------------------------------------------------*/
void
stfp_preconfig(int point, int order, ulong argument, ulong event_arg)
{
   struct controller_config ctlr_register;
   struct controller_config *ctlr_register_ptr = &ctlr_register;
   int i;

   ctlr_register_ptr->revision = CTLR_CONFIG_REVISION;
   strcpy(ctlr_register_ptr->subsystem_name,mcfgname);
   strcpy(ctlr_register_ptr->bus_name,DRIVER_WILDNAME);
   ctlr_register_ptr->devdriver = &stfpdriver;

   for(i=0;i<stfp_num_installed;i++) {
      preconfig_status = create_controller_struct(ctlr_register_ptr);
      if (preconfig_status != ESUCCESS) {
         uprintf("stfp_preconfig: could not create controller struct\n");
         uprintf("preconfig: create_controller_struct #%d Failed!\n", i);
         return;
      }
      uprintf("preconfig: create_controller_struct #%d Done!\n", i);
   }

   return;
}

/*---------------------------------------------------------
* post configure
* 1 - Add Device Switch Table.
*--------------------------------------------------------*/
void
stfp_postconfig(int point, int order, ulong argument, ulong event_arg)
{
   if (preconfig_status != ESUCCESS) return;

   majnum = devsw_add(mcfgname,MAJOR_INSTANCE, majnum,
                      &stfp_devsw_entry);

   if(majnum == ENODEV) {
      if (stfp_developer_debug)
         uprintf("stfp_postconfig: call to devsw_add failed\n");
      return;
   }

   already_configured = TRUE;
   
   if(stfp_developer_debug) {
      uprintf("stfp_postconfig:major # = %d\n",majnum);
      uprintf("STFP driver V.%s configured\n",stfp_version);
   }
   
   return;
}

/*---------------------------------------------------------
* Attach controller
* return values: none
*--------------------------------------------------------*/
int stfp_cattach(struct controller *ctlr)
{
   struct stfp_softc *scp = stfp_softc[ctlr->ctlr_num];

   if(stfp_developer_debug)
	uprintf("stfp_cattach: Entry!\n");

   /* initialize the spin lock */

   STFP_INIT_LOCK(&scp->stfpLock);
   STFP_INIT_LOCK(&scp->sc_intLock);

   return(1);
}


/*---------------------------------------------------------
* Ctlr_unattach interface
* 1 - UnDo what was done in the Probe interface.
*--------------------------------------------------------*/
int stfp_ctlr_unattach(struct bus *busptr, struct controller *ctlr)
{
   int unit = ctlr->ctlr_num;
   struct stfp_softc *scp = stfp_softc[unit];
   int status;

   if (stfp_developer_debug)
      uprintf("stfp_ctlr_unattach: ctlr # = %d\n",unit);
   
   if (unit < 0 || unit >= stfp_num_probed)
      return ENODEV;
   
   if(driver_cfg_state == SUBSYSTEM_STATICALLY_CONFIGURED)
      return ENODEV; /* static drivers need not proceed */
   
   /* free softc structure for this controller only */
   FREE(scp,M_DEVBUF);
   
   return ESUCCESS;
}


/*---------------------------------------------------------
* Probe interface
* called once for each controller
* 1 - Check to see if hardware exists
* 2 - allocate softc structure
* returns: 0 for fail, 1 for success
*--------------------------------------------------------*/
stfp_probe(io_handle_t bus_io_handle, struct controller *ctlr)
{
  ihandler_t handler;
  int unit = ctlr->ctlr_num;
  struct stfp_softc *scp;
  int status = 0, detect_hardware = 0, i = 0;
  struct pci_config_hdr *pci_cfg_hdr = (struct pci_config_hdr *)bus_io_handle;

/* ===> */

  unsigned int mem_cfg;
  unsigned char *stfp_addr_ptr;

  int datain, time0, time1;
  unsigned char   dpr_value;
  unsigned char   dpr_cmd;

  char  temp_char;

  struct handler_intr_info stfp_intr_info;
  ihandler_t stfp_ihandle;

/* <=== */


  if(stfp_developer_debug)
    uprintf("stfp_probe: Entry!\n");

  /* allocate softc, better to ZERO-fill! */   

  MALLOC(stfp_softc[unit], struct stfp_softc *, 
         sizeof(struct stfp_softc), M_DEVBUF, M_NOWAIT | M_ZERO);

  if (stfp_softc[unit] == (struct stfp_softc *)NULL) {
     uprintf("stfp_probe: Unable to malloc stfp_softc[%d]\n",unit);
     return 0;
  }

  scp = stfp_softc[unit];
  stfp_info[unit] = ctlr;
  scp->is_open = 0;

  /* handle bus differences */

  switch(ctlr->bus_hd->bus_type){
   case BUS_PCI:
     scp->iobase = pci_cfg_hdr->bar0;
     break;
   default:
     uprintf("stfp_probe: Unsupported bus type: %d\n",
            ctlr->bus_hd->bus_type);
  }

  /* 
   * Detect the controller and device hardware here 
   * status = detect_hardware;
   */   

  detect_hardware = 1;	/* Now the h/w is probed!? */

/* ===> */

/* Ensure that we've found the device                            */

   if (!(pci_cfg_hdr->command & CMD_MEM_SPACE))
   {
     if (stfp_developer_debug)
         printf ("probe: %s%d: not found by POST code\n", STFP_NAME, unit);
      return (0);
   }

    if(unit > MAX_STFP){
      printf("failed: stfp unit greater than max_stfp");
      return(0);
    }

   scp->sc_unit = unit;
   scp->sc_ctlr = ctlr;

/* init device area */

   bzero((char *) scp, sizeof(struct stfp_softc));
   scp->sc_unit = unit;
   scp->sc_ctlr = ctlr;
   scp->enablesignal = DISABLE_PSIGNAL;

   /* make registers accessible */

   scp->Regshandle = pci_cfg_hdr->bar0;
   scp->Dprhandle  = romsize + pci_cfg_hdr->bar1;

   printf( "%s%d:  Regs base address = 0x%x\n", 
	STFP_NAME, unit, scp->Regshandle);
   printf( "%s%d:  Dual Port RAM base address = 0x%x\n", 
	STFP_NAME, unit, scp->Dprhandle);

/* Read the Dual Port RAM offset Locations from top of DPRAM  */

        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 1), &dpr_value );
        scp->pIn = 0xff & dpr_value;
        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 2), &dpr_value );
        scp->pIn = scp->pIn | (0xff00 & (dpr_value << 8));
        printf("scp->pIn Offset = %x\n",scp->pIn);

        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 3), &dpr_value );
        scp->pOut = 0x00ff & dpr_value;
        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 4), &dpr_value );
        scp->pOut = scp->pOut | (0xff00 & (dpr_value << 8));
        printf("scp->pOut Offset = %x\n",scp->pOut);

        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 5), &dpr_value );
        scp->pGps = 0x00000000000000ff & dpr_value;
        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 6), &dpr_value );
        scp->pGps = scp->pGps | (0x000000000000ff00 & (dpr_value << 8));
        printf("scp->pGps Offset = %x\n",scp->pGps);

        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 7), &dpr_value );
        scp->pYear = 0x00000000000000ff & dpr_value;
        STFP_RD_BUS_D8(scp->Dprhandle + (TOP_DPRAM - 8), &dpr_value );
        scp->pYear = scp->pYear | (0x000000000000ff00 & (dpr_value << 8));
        printf("scp->pYear Offset = %x\n",scp->pYear);

/* Read time three times */

   STFP_RD_BUS_D32(scp->Regshandle + TIMEREQ, &datain);
   STFP_RD_BUS_D32(scp->Regshandle + TIME0, &time0);
   STFP_RD_BUS_D32(scp->Regshandle + TIME1, &time1);

   printf("time0 = %08x time1 = %08x\n",time0,time1);

   STFP_RD_BUS_D32(scp->Regshandle + TIMEREQ, &datain);
   STFP_RD_BUS_D32(scp->Regshandle + TIME0, &time0);
   STFP_RD_BUS_D32(scp->Regshandle + TIME1, &time1);

   printf("time0 = %08x time1 = %08x\n",time0,time1);

   STFP_RD_BUS_D32(scp->Regshandle + TIMEREQ, &datain);
   STFP_RD_BUS_D32(scp->Regshandle + TIME0, &time0);
   STFP_RD_BUS_D32(scp->Regshandle + TIME1, &time1);

   printf("time0 = %08x time1 = %08x\n",time0,time1);

   if (scp->pIn != 0x102)
   {
/*
      cmn_err(CE_CONT, "635PCI Did Not Boot Correctly - Try system warm reboot\n");
*/
      printf("635PCI Did Not Boot Correctly - Try system warm reboot\n");
      return (0);
   }

   if (stfp_developer_debug)
      printf ("%s%d probe: calling init\n", STFP_NAME, unit);

   /* register interrupts */

   stfp_intr_info.configuration_st = (caddr_t)ctlr;
   stfp_intr_info.intr = stfp_intr;
   stfp_intr_info.param = (caddr_t)unit;
   stfp_intr_info.config_type = (CONTROLLER_CONFIG_TYPE | SHARED_INTR_CAPABLE);
   stfp_ihandle.ih_bus = ctlr->bus_hd;
   stfp_ihandle.ih_bus_info = (char *) &stfp_intr_info;

   scp->sc_hid = handler_add(&stfp_ihandle);
   if (scp->sc_hid == (ihandler_id_t *)(NULL))
   {
    printf ("stfp_probe:  %s%d:  \
	interrupt handler add failed\n", STFP_NAME, unit);
    return (0);
   }
   printf ("stfp_probe:  interrupt handler add successful\n");

   if (handler_enable(scp->sc_hid) != 0)
   {
      printf ("stfp_probe:  interrupt handler enable failed\n");
      handler_del(scp->sc_hid);
      return (0);
   }

   printf ("stfp_probe:  interrupt handler enable successful\n");
   printf("stfpprobe:successful\n");
/*
   cmn_err(CE_CONT, "Bancomm Timing BC635/637 PCI TimeCode Rec/Gen.  Revision x.x\n");
   cmn_err(CE_CONT, "Copyright (c) 1999 Northrop Grumman, Inc.  all rights reserved\n");
*/

/* <=== */


  status = detect_hardware;
  if (!status) {
     uprintf("stfp_probe: Unable to detect controller! \n");
     return 0;
  }

  if (stfp_developer_debug)
     uprintf("stfp_probe: Successful!\n");
  
  stfp_num_probed++;

  return 1;
}

/*---------------------------------------------------------
* Open
* return: ESUCCESS on success. ENODEV,EBUSY,ENXIO on fail.
*--------------------------------------------------------*/
int stfp_open(dev_t dev, int flag, int format)
{

   int unit = minor(dev);
   struct stfp_softc *scp = stfp_softc[unit];
   struct controller *ctlr = stfp_info[unit];

   if (stfp_developer_debug)
      uprintf("stfp_open: Entry! \n");
   
   /* only allow one open device */
   if (scp->is_open)
      return EBUSY;

   if (scp->sc_openf == ST_OPEN)
   {
      if (stfp_developer_debug)
         printf("stfp_open: %s%d: device already opened\n", MAX_STFP, unit);
      return (EBUSY);
   }

   /* minor device must exist */
   if (unit >= MAX_STFP) {
      if (stfp_developer_debug)
         printf("stfp_open: %s%d: no such unit (max=%d)\n", \
		STFP_NAME, unit, MAX_STFP);
      return (ENODEV);
   }  

   if ((ctlr != 0) && (ctlr->alive & ALV_ALIVE)) {
      scp->sc_openf = ST_OPEN;
      scp->interrupt_count = 0; /* Init int. count  */
      scp->signal = 0;          /* Init no signal   */
      scp->waiting= FALSE;      /* Init not waiting */

      scp->is_open = 1;

      if (stfp_developer_debug)
         printf("stfp_open: %s%d: device opened\n", STFP_NAME, unit);
      return (ESUCCESS);
   } else {
      if (stfp_developer_debug)
      printf("stfp_open: %s%d: device not found\n", STFP_NAME, unit);
      return (ENXIO);
   }

   return ESUCCESS;
}

/*---------------------------------------------------------
* Close
* return: ESUCCESS on success. ENODEV,EBUSY,ENXIO on fail.
*--------------------------------------------------------*/
int stfp_close(dev_t dev, int flag, int format)
{

  int unit = minor(dev);
  struct stfp_softc *scp = stfp_softc[unit];
  struct controller *ctlr = stfp_info[unit];
  register io_handle_t reg = (io_handle_t) ctlr->addr;

  if (stfp_developer_debug)
     uprintf("stfp_close: Entry! \n");

  /* turn off the interrupt handler signal  */
  
  scp->enablesignal = DISABLE_PSIGNAL;

  /* turn off the open flag */
  scp->sc_openf = ST_CLOSE;
  
  scp->is_open = 0;
   
  return ESUCCESS;
}

/*---------------------------------------------------------
* Read -
* return: ESUCCESS on success. 
*         fail: An error number from errno.h
*--------------------------------------------------------*/
int 
stfp_read(dev_t dev, struct uio *uioptr, int flag)
{

   int unit = minor(dev);
   struct stfp_softc *scp = stfp_softc[unit];
   int i,status;
   int size=uioptr->uio_resid;
   int dma_size;

   if (stfp_developer_debug)
      uprintf("stfp_read:entry!\n");

   uioDump( scp, uioptr, "stfpread" );

   return ESUCCESS;

}

int 
stfp_write(dev_t dev, struct uio *uioptr, int flag)
{

   int unit = minor(dev);
   struct stfp_softc *scp = stfp_softc[unit];
   int i,status, n, dma_size;
   int size=uioptr->uio_resid;

   if (stfp_developer_debug)
      uprintf("stfp_write:entry\n");

   uioDump( scp, uioptr, "stfpwrite" );

   if ((n = uioptr->uio_iov->iov_len) > STFP_MAX_WRITE)
      return (EFBIG);
    else
      status = 0;

   /* copy the data from the user's buffer */


   return ESUCCESS;

}

/*---------------------------------------------------------
* IOCTL
* return: ESUCCESS on success. 
*         fail: An error number from errno.h
*--------------------------------------------------------*/
int 
stfp_ioctl(dev_t dev, unsigned int cmd, caddr_t data, int flag)
{

   int unit = minor(dev);
   struct stfp_softc *scp = stfp_softc[unit];
   struct stfp_ioc_type *stfpioct = (struct stfp_ioc_type *)data;

   unsigned long  datain;
   static stfp_time_t     Tmarg;
   u_char   *getdataptr;

   int      get_data_len;
   int      status;
   int      argument;
   u_int    time;
   short    i;

   int *res;
   int err;
   long dpr_cmd, dpr_data;
   long dpr_value;

   status = 0;
   err = 0;  /* Set error flag to pass  */


   if (stfp_developer_debug)
      uprintf("stfp_ioctl:entry\n");
   
   switch(cmd) {

   /* Snapshot the time registers */
      case TIMEREQUEST:
      {
         STFP_REG_PUTL(TIMEREQ, 0);
      }
      break;

   /* Snapshot the event registers */
      case EVENTREQUEST:
      {
         STFP_REG_PUTL(EVENTREQ, 0);
       }
      break;

   /* Read the time registers and format time */
      case RDTIME:
      {
        TmPtr Tm  = (TmPtr) data;
        rdTime(unit, TIME0, scp->pYear, Tm);

      }
      break;

   /* Read the event registers and format time */
      case RDEVENT:
      {
        TmPtr Tm  = (TmPtr) data;
        rdTime(unit, EVENT0, scp->pYear, Tm);

      }
      break;

   /* Snapshot time registers, Read the time registers and format the time */
      case RDTIMEREQ:
      {
        TmPtr Tm  = (TmPtr) data;
        STFP_REG_PUTL(TIMEREQ, 0);
        rdTime(unit, TIME0, scp->pYear, Tm);
      }
      break;

   /* Snapshot event registers, Read the event registers and format the time */
      case RDEVENTREQ:
      {
        TmPtr Tm  = (TmPtr) data;
        STFP_REG_PUTL(EVENTREQ, 0);
        rdTime(unit, EVENT0, scp->pYear, Tm);
      }
      break;

/*    Setup the bc635PCI card timebase parameters */
      case CONTROLTIMEBASE:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: CONTROLTIMEBASE entry\n");

          switch (*(u_short*)data)
         {
            case CLOCK_INTERNAL:
               STFP_DP_PUTB(scp->pIn, COM_SELCLOCK);
               STFP_DP_PUTB(scp->pIn+1, SELCLOCK_INT);
               status = sendCommand(unit);
            break;
            case CLOCK_EXTERNAL:
               STFP_DP_PUTB(scp->pIn, COM_SELCLOCK);
               STFP_DP_PUTB(scp->pIn+1, SELCLOCK_EXT);
               status = sendCommand(unit);
            break;
            case JAMSYNC_ENABLE:
               STFP_DP_PUTB(scp->pIn, COM_JAMSYNCENABLE);
               STFP_DP_PUTB(scp->pIn+1, JAMSYNC_EN);
               status = sendCommand(unit);
            break;
            case JAMSYNC_DISABLE:
               STFP_DP_PUTB(scp->pIn, COM_JAMSYNCENABLE);
               STFP_DP_PUTB(scp->pIn+1, JAMSYNC_DIS);
               status = sendCommand(unit);
            break;
            case DISCIPLINE_ENABLE:
               STFP_DP_PUTB(scp->pIn, COM_DISCENABLE);
               STFP_DP_PUTB(scp->pIn+1, DISC_EN);
               status = sendCommand(unit);
            break;
            case DISCIPLINE_DISABLE:
               STFP_DP_PUTB(scp->pIn, COM_DISCENABLE);
               STFP_DP_PUTB(scp->pIn+1, DISC_DIS);
               status = sendCommand(unit);
            break;
            case FORCE_JAMSYNC:
               STFP_DP_PUTB(scp->pIn, COM_FORCEJAMSYNC);
               status = sendCommand(unit);
            break;
            case CLEAR_USEC:
            break;
            default:
               status = EINVAL;
            break;
         }
      }
      break;

/*    Set DAC (not tested) */
      case SETDAC:
      {
         STFP_DP_PUTB(scp->pIn, COM_SETDAC);
         STFP_DP_PUTB(scp->pIn+1, (*(unsigned short*)data >> 8) & 0xff);
         STFP_DP_PUTB(scp->pIn+2, (*(unsigned short*)data) & 0xff);
         status = sendCommand(unit);
      }
      break;

/*    Set gain (not tested) */
      case SETDISCGAIN:
      {
         STFP_DP_PUTB(scp->pIn, COM_SETDISCGAIN);
         STFP_DP_PUTB(scp->pIn+1, (*(short*)data >> 8) & 0xff);
         STFP_DP_PUTB(scp->pIn+2, (*(short*)data) & 0xff);
         status = sendCommand(unit);
      }
      break;

/*    Set timing offset (not tested) */
      case SETTIMINGOFFSET:
      {
         STFP_DP_PUTB(scp->pIn, COM_SETTIMINGOFFSET);
         STFP_DP_PUTB(scp->pIn+1, (*(int*)data >> 24) & 0xff);
         STFP_DP_PUTB(scp->pIn+2, (*(int*)data >> 16) & 0xff);
         STFP_DP_PUTB(scp->pIn+3, (*(int*)data >>  8) & 0xff);
         STFP_DP_PUTB(scp->pIn+4, (*(int*)data)       & 0xff);
         status = sendCommand(unit);
      }
      break;

/*    Set timing mode . GPS mode is not functional */
      case SELTIMINGMODE:
      {
         STFP_WRT_BUS_D8(scp->Dprhandle + (scp->pIn + 0), COM_SETTIMINGMODE);
         STFP_WRT_BUS_D8(scp->Dprhandle + (scp->pIn + 1), (*(unsigned short*)data));

         status = sendCommand(unit);
      if (stfp_developer_debug)printf("stfp_ioctl: SELTIMINGMODE = %x\n",*(unsigned char*)data);
      }
      break;


/*    Set timing format . Binary is only option */
      case SELTIMEFORMAT:
      {
         STFP_DP_PUTB(scp->pIn, COM_SELTIMEFORMAT);
         STFP_DP_PUTB(scp->pIn+1, *(unsigned char*)data);
         status = sendCommand(unit);
      if (stfp_developer_debug)printf("stfp_ioctl: SELTIMEFORMAT = %x\n",*(unsigned char*)data);
      }
      break;

/*    Select timecode format.  i.e. IRIG-B */
      case SELTCFORMAT:
      {
         STFP_DP_PUTB(scp->pIn, COM_SETTCFORMAT);
         STFP_DP_PUTB(scp->pIn+1, *(u_char*)data);
         status = sendCommand(unit);
      if (stfp_developer_debug)printf("stfp_ioctl: SELTCFORMAT = %x\n",*(unsigned char*)data);
      }
      break;

/*    Select time code mode */
      case SELTCMOD:
      {
         STFP_DP_PUTB(scp->pIn, COM_SETTCMOD);
         STFP_DP_PUTB(scp->pIn+1, *(u_char*)data);
         status = sendCommand(unit);
      if (stfp_developer_debug)printf("stfp_ioctl: SELTCMOD = %x\n",*(unsigned char*)data);
      }
      break;

/*    Set periodic heartbeat. Rate = 10,000,000 / (n1 * n1) */
/*    Set periodic SYNC to allow sync to 1PPS               */
      case SETPERIODIC:
      {
         PerPtr Per  = (PerPtr) data;

         STFP_DP_PUTB(scp->pIn, COM_SETPERIODIC);
         STFP_DP_PUTB(scp->pIn+1, Per->sync1pps);
         STFP_DP_PUTB(scp->pIn+2, (uchar)((Per->n1 >> 8) & 0x00ff));
         STFP_DP_PUTB(scp->pIn+3, (uchar) (Per->n1 & 0x00ff));
         STFP_DP_PUTB(scp->pIn+4, (uchar)((Per->n1 >> 8) & 0x00ff));
         STFP_DP_PUTB(scp->pIn+5, (uchar) (Per->n2 & 0x00ff));

         status = sendCommand(unit);
      }
      break;

/*    Set CONTROL register.  Not fully tested */
      case CONTROLEVENT:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: CONTROLEVENT = %x\n",*(unsigned char*)data);
         switch (*(u_char*)data)
         {
            case EVENT_RISING:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_EVSENSE);
            break;
            case EVENT_FALLING:
               STFP_REG_OREQUAL(CONTROL, CTRL_EVSENSE);
            break;
            case CAPLOCK_ENABLE:
               STFP_REG_OREQUAL(CONTROL, CTRL_LOCKEN);
            break;
            case CAPLOCK_DISABLE:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_LOCKEN);
            break;
            case EVCAP_ENABLE:
               STFP_REG_OREQUAL(CONTROL, CTRL_EVENTEN);
            break;
            case EVCAP_DISABLE:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_EVENTEN);
            break;
            case EVCAP_PERIODIC:
               STFP_REG_OREQUAL(CONTROL, CTRL_EVSOURCE);
            break;
            case EVCAP_EVENT:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_EVSOURCE);
            break;
            case CLEAR_CONTROL:
               STFP_REG_PUTL(CONTROL, 0);
            break;
            case READ_CONTROL:
             STFP_RD_BUS_D32(scp->Regshandle + CONTROL, &datain);
              printf("CONTROL = %x\n",datain);
            break;
            case SET_CONTROL:
               STFP_REG_OREQUAL(CONTROL, 10);
            break;
            default:
               status = EINVAL;
               break;
         }
      }
      break;

/*    Control Strobe.  Not tested */
      case CONTROLSTROBE:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: CONTROLSTROBE = %x\n",*(unsigned char*)data);
         switch (*(u_char*)data)
         {
            case STROBE_SECUS:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_STRMODE);
            break;
            case STROBE_USONLY:
               STFP_REG_OREQUAL(CONTROL, CTRL_STRMODE);
            break;
            case STROBE_ENABLE:
               STFP_REG_OREQUAL(CONTROL, CTRL_STREN);
            break;
            case STROBE_DISABLE:
               STFP_REG_ANDEQUAL(CONTROL, ~CTRL_STREN);
            break;
            default:
               status = EINVAL; break;
         }
      }
      break;

/*    Select connector output frequency.  Not tested  */
      case SELFREQUENCYOUT:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: SELFREQUENCYOUT = %x\n",*(unsigned char*)data)
;
         switch (*(u_char*)data)
         {
            case FREQ_10MHZ:
               STFP_REG_PUTL(CONTROL,
                             (STFPRegGetL(scp->Regshandle+CONTROL)
                              & ~CTRL_FMASK) | CTRL_F10);

            break;
            case FREQ_5MHZ:
               STFP_REG_PUTL(CONTROL,
                             (STFPRegGetL(scp->Regshandle+CONTROL)
                              & ~CTRL_FMASK) | CTRL_F5);

            break;
            case FREQ_1MHZ:
               STFP_REG_PUTL(CONTROL,
                             (STFPRegGetL(scp->Regshandle+CONTROL) &
                              ~CTRL_FMASK) | CTRL_F1);

            break;
            default:
               status = EINVAL; break;
         }
      }
      break;

/*    Get data (from Dual Port Memory) command.  Call GETDATA for result */
      case COMMANDGETDATA:
      {
        scp->getdata.arg = *(uchar*)data;
        if (stfp_developer_debug)printf("stfp_ioctl: COMMANDGETDATA = %x\n",scp->getdata.arg);

        STFP_DP_PUTB(scp->pIn, COM_GETDATA);
        STFP_DP_PUTB(scp->pIn+1, scp->getdata.arg);

        if(status = sendCommand(unit) != 0)
         {
          printf("Failed GETDATA. Arg = \n",scp->getdata.arg);
          break;
         }
       }
       break;


/*    Get data from Dual Port Memory.  Call COMMANDGETDATA for setup */
      case GETDATA:
      {
        scp->getdata.arg = *(uchar*)data;
        if (stfp_developer_debug)printf("stfp_ioctl: GETDATA = %x\n",scp->getdata.arg);

         /* get the number of bytes to be transferred */

          get_data_len = 0;

        switch(scp->getdata.arg)
         {
            case GETDATA_MODE:
               get_data_len=sizeof(scp->getdata.data.tmode);
            break;
            case GETDATA_TFORMAT:
               get_data_len=sizeof(scp->getdata.data.tformat);
            break;
            case GETDATA_YEAR:
               get_data_len=sizeof(scp->getdata.data.year);
            break;
            case GETDATA_PER:
               get_data_len=sizeof(scp->getdata.data.period);
            break;
            case GETDATA_TCFMT:
               get_data_len=sizeof(scp->getdata.data.tcformat);
            break;
            case GETDATA_TCMOD:
               get_data_len=sizeof(scp->getdata.data.tcmod);
            break;
            case GETDATA_TOFF:
               get_data_len=sizeof(scp->getdata.data.offset);
            break;
            case GETDATA_UTCINFO:
               get_data_len=sizeof(scp->getdata.data.utc);
            break;
            case GETDATA_FWVER:
               get_data_len=sizeof(scp->getdata.data.version);
            break;
            case GETDATA_CLKSRC:
               get_data_len=sizeof(scp->getdata.data.clksrc);
            break;

            case GETDATA_JAMSC:
               get_data_len=sizeof(scp->getdata.data.jsctl);
            break;
            case GETDATA_OSCCTL:
               get_data_len=sizeof(scp->getdata.data.ocsctl);
            break;
            case GETDATA_DAVAL:
               get_data_len=sizeof(scp->getdata.data.davalue);
            break;
            default:
               get_data_len=0;
               status=EFAULT;
               break;


         }
         /*  Only return data */

         STFP_RD_BUS_D8(scp->Dprhandle + scp->pOut, &dpr_value );
         if(dpr_value == scp->getdata.arg){
                 for(i=0; i<get_data_len+1; i++){
                    STFP_RD_BUS_D8(scp->Dprhandle + scp->pOut+i, &dpr_value );
                    *(uchar*)(data+i)  = (unsigned char)dpr_value;
                  if (stfp_developer_debug)printf("GETDATA = %02x i = %d\n",dpr_value,i);
                  }
        }else{if (stfp_developer_debug)printf("GETDATA failed\n");}
      }
      break;

/*    The Tv options allow a different formatting option for time */
/*    Modify rdTimeTv as desired */

      case RDTIMETV:
      {
         TvPtr Tv  = (TvPtr) data;
         rdTimeTv(unit, TIME0, Tv);
      }
      break;

      case RDEVENTTV:
      {
         TvPtr Tv  = (TvPtr) data;
         rdTimeTv(unit, EVENT0, Tv);
      }
      break;

      case RDTIMETVREQ:
      {
         TvPtr Tv  = (TvPtr) data;
         STFP_REG_PUTL(TIMEREQ, 0);
         rdTimeTv(unit, TIME0, Tv);
      }
      break;

      case RDEVENTTVREQ:
      {
         TvPtr Tv  = (TvPtr) data;
         STFP_REG_PUTL(EVENTREQ, 0);
         rdTimeTv(unit, EVENT0, Tv);
         status = EFAULT;
      }
      break;


/*    Allow interrupts */
      case CAPUNLOCK:
      {
         STFP_REG_PUTL(RELEASE_CAP, 0);
      }
      break;

/*    Read INTSTAT register */
      case RDINTSTAT:
      {
         *(unsigned int*)data = STFPRegGetL(scp->Regshandle+INTSTAT) & INT_BITMASK;
      if (stfp_developer_debug)printf("stfp_ioctl: RDINTSTAT = %x\n",*(unsigned int*)data);
      }
      break;

/*    Clear INTSTAT register bits */
      case CLRINTSTAT:
      {
         STFP_REG_PUTL(INTSTAT, *(unsigned int*)data & ~scp->intMaskSig);
      if (stfp_developer_debug)printf("stfp_ioctl: CLRINTSTAT = %x\n",*(unsigned int*)data);
      }

      break;

/*    Set strobe time.  Not tested */
      case WRSTROBE:
      {
         TmPtr Tm  = (TmPtr) data;

         STFP_REG_ANDEQUAL(CONTROL, ~CTRL_STREN);

         STFP_REG_PUTL(EVENT0+1,
                       ((Tm->tm.tm_hour << 16) + (Tm->tm.tm_min << 8)
                         + Tm->tm.tm_sec) & 0xffffff);

         STFP_REG_PUTL(EVENT0, Tm->tm_usec & 0xfffff);
         STFP_REG_OREQUAL(CONTROL, CTRL_STREN);
      }
      break;

      case WRSTROBETV:
      {
         TvPtr Tv  = (TvPtr) data;

         STFP_REG_ANDEQUAL(CONTROL, ~CTRL_STREN);
         STFP_REG_PUTL(EVENT0+1, Tv->tv.tm_sec);
         STFP_REG_PUTL(EVENT0, Tv->tv_usec & 0xfffff);
         STFP_REG_OREQUAL(CONTROL, CTRL_STREN);
      }
      break;

/*    Initialize time.  Not tested */
      case SETTIME:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: SETTIME = %x\n",*(unsigned int*)data);
         STFP_DP_PUTB(scp->pIn, COM_SETTIME);
         STFP_DP_PUTB(scp->pIn+1, (*(unsigned int*)data >> 24) & 0x000000ff);
         STFP_DP_PUTB(scp->pIn+2, (*(unsigned int*)data >> 16) & 0x000000ff);
         STFP_DP_PUTB(scp->pIn+3, (*(unsigned int*)data >> 8) & 0x000000ff);
         STFP_DP_PUTB(scp->pIn+4,  *(unsigned int*)data & 0x000000ff);
         status = sendCommand(unit);

      }
      break;

      case SETTIMETV: /* Get time from tv structure */
      {
         TvPtr Tv  = (TvPtr) data;
      if (stfp_developer_debug)printf("stfp_ioctl: SETTIMETV = %x\n",*(unsigned int*)data);

         STFP_DP_PUTB(scp->pIn, COM_SETTIME);
         STFP_DP_PUTB(scp->pIn+1, (Tv->tv.tm_sec >> 24) & 0xff);
         STFP_DP_PUTB(scp->pIn+2, (Tv->tv.tm_sec >> 16) & 0xff);
         STFP_DP_PUTB(scp->pIn+3, (Tv->tv.tm_sec >> 8) & 0xff);
         STFP_DP_PUTB(scp->pIn+4, Tv->tv.tm_sec & 0xff);
         status = sendCommand(unit);


      }
      break;

/*   Initialize year.  Card incorporates year data into time registers.  */
/*   Delay a bit after this call */
     case SETYEAR:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: SETYEAR = %x\n",*(unsigned short*)data);
         STFP_WRT_BUS_D8(scp->Dprhandle + (scp->pIn + 0), COM_SETYEAR);
         STFP_WRT_BUS_D8(scp->Dprhandle + (scp->pIn + 1), (*(unsigned short*)data >> 8) & 0xff);
         STFP_WRT_BUS_D8(scp->Dprhandle + (scp->pIn + 2),  *(unsigned short*)data & 0xff);
         status = sendCommand(unit);
      }
      break;

/*    Retrieve the Requested RT signal associated with this driver */
      case SETRTSIGNAL:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: SETRTSIGNAL = %x\n",*(unsigned int*)data);

         scp->signal = *(unsigned int*)data;

/*    Get a pointer to the calling process.  Delete this? */

       if (scp->intMaskSig)scp->proc_p = u.u_procp;

      }
      break;

/*  Set the MASK register interrupt bit */
      case SETINTSIGNAL:
      {
      if (stfp_developer_debug)printf("stfp_ioctl: SETINTSIGNAL = %x\n",*(unsigned int*)data);

/*  Retrieve the interrupt type request */

         scp->intMaskSig = *(unsigned int*)data & INT_BITMASK;

/*  Clear any pending interrupt bit in INTSTAT register */

         STFP_REG_PUTL(INTSTAT, scp->intMaskSig);

/*  Set the requested interrupt enable bit in the MASK register */

         STFP_REG_PUTL(MASK,
                       (STFPRegGetL(scp->Regshandle+MASK)
                        & INT_PACKET) | scp->intMaskSig);
      }
      break;

/*    Read back the interrupt which was serviced last by the int handler */
      case RDINTSIGNAL:
      {
         argument = scp->intServiced & INT_BITMASK;
         scp->intServiced = 0;
        *(unsigned int *)data = argument;
      if (stfp_developer_debug)printf("stfp_ioctl: RDINTSIGNAL = %x\n",*(unsigned int*)data);
      }
      break;

/*    Enable the signal from the int handler */
      case SIGNALCONTROL:
      {
       scp->enablesignal = *(char*)data;
      }
      break;

/*    Retreive the time when the last interrupt occurred */
      case RDINTTIME:
      {
       *(stfp_time_t *)data = scp->IntTime;
      }
      break;

/*    Reset the Card */
      case RESETTFP:
      {
       dpr_cmd = COM_RESET;
       STFP_WRT_BUS_D8(scp->Dprhandle + scp->pIn, COM_RESET);
       status = sendCommand(unit);
      }
      break;


    default:
      uprintf("stfp_ioctl: Undefined command\n");
   }

   return ESUCCESS;

}


/******************* Supporting Interfacaes ************************/


/*******************************************************************
 *                                                                 *
 *  It is important for user to understand the following about     *
 *  how mmap() works at the operating system level.                *
 *                                                                 *
 *  - The mmap routine is invoked by the kernel as a result of     *
 *    an application calling the mmap(2) system call. The mmap     *
 *    routine makes sure that the specified offset into the memory *
 *    mapped device's memory is valid.                             *
 *                                                                 *
 *    mmap returns the page frame number corresponding to the      *
 *    page at the specified offset.                                *
 *                                                                 *
 *  - mmap() is used to map physical locations on an external      *
 *    device to virtual addresses in the calling process address   *
 *    space.                                                       *
 *                                                                 *
 *  - The operating system will create a memory object in the      *
 *    calling process' virtual memory map and mark all the pages   *
 *    as "invalid".  This will cause a page fault to occur for     *
 *    each page as they are accessed.                              *
 *                                                                 *
 *  - The operating system will then call the device driver mmap() *
 *    routine for each page fault to perform the address           *
 *    translation.                                                 *
 *                                                                 *
 *  - Once the translation has been done for a page, that page is  *
 *    then accessed directly by the process and the driver is no   *
 *    longer involved.                                             *
 *                                                                 *
 *******************************************************************/

int
stfp_mmap(dev_t dev, off_t off, int prot)
{
  register int unit = minor(dev);
  struct stfp_softc *scp = stfp_softc[unit];

  uprintf("stfp_mmap: Entry! off %d prot %d\n", off, prot);

  return(ESUCCESS);

}


/*******************************************************************/
/* stfp_intr - low level interrupt handler                         */
/*******************************************************************/

int stfp_intr(register int unit )
{
   register struct controller *ctlr = stfp_info[unit];
   volatile int pending;
   int prevSpl;                          /* Previous CPU priority level  */
   struct stfp_softc *scp = stfp_softc[unit];   
   TmPtr Tm  = &scp->IntTime;

/*  Determine which interrupt type happened */

    pending = STFPRegGetL(scp->Regshandle+MASK) &
         STFPRegGetL(scp->Regshandle+INTSTAT) & INT_BITMASK;

/*  Claim or reject interrupt check */

   if(pending == 0)
   {
      if (stfp_developer_debug)
         printf("stfp_intr: Interrupt UNCLAIMED!!!\n");
      return(INTR_NOT_SERVICED);
   }

/* Acquire the device lock */

      if (stfp_developer_debug)
    {
        printf( "%s%d: locking interface stfp_intr\n",
                       ctlr->ctlr_name,
                       ctlr->ctlr_num);
    }

    STFP_LOCK( &scp->sc_intLock );

/*   Get Time of interrupt */

   STFP_REG_PUTL(TIMEREQ, 0);
   rdTime(unit, EVENT0, scp->pYear, Tm);

/*  Clear the Pending Interrupt Bit */

   scp->intServiced |= pending & scp->intMaskSig;
   STFP_REG_PUTL(INTSTAT,scp->intServiced);

/*   Increment the interrupt counter */

   scp->interrupt_count++;
   if (stfp_developer_debug)printf("stfp_intr: entry = %d %x\n",scp->interrupt_count,pending);

/*  Send the signal to the parent process */

   if(scp->enablesignal == 1)
    psignal( scp->proc_p, scp->signal );

   STFP_REG_PUTL(RELEASE_CAP, 0);

   /* Restore the device lock */

   STFP_UNLOCK( &scp->sc_intLock );

   return(INTR_SERVICED);

}


/*------------------------ USER FUNCTIONS -----------------------------------*/

/*****************************************************************************
* rdTimeTv - read time of day from the board in binary seconds format ***************
*  Change this if any special formatting is required
******************************************************************************/

void rdTimeTv(int unit, uint pTmReg, stfp_timeval_t *pTmS)
{
   unsigned int time0, time1, datain;

   struct stfp_softc *scp = stfp_softc[unit];

   STFP_RD_BUS_D32(scp->Regshandle + TIMEREQ, &datain);
   STFP_RD_BUS_D32(scp->Regshandle + TIME0, &time0);
   STFP_RD_BUS_D32(scp->Regshandle + TIME1, &time1);

   pTmS->tv.tm_yday =  time1 / 86400;

   pTmS->tv.tm_hour =  (time1 - pTmS->tv.tm_yday*86400)    / 3600;
 
   pTmS->tv.tm_min  =  (time1 - pTmS->tv.tm_yday*86400 - 
                                pTmS->tv.tm_hour*3600 )    / 60;

   pTmS->tv.tm_sec  =  (time1 - pTmS->tv.tm_yday*86400 - 
                                pTmS->tv.tm_hour*3600  - 
                                pTmS->tv.tm_min*60);
  
   pTmS->tv_usec       =  time0 & 0xfffff;
   pTmS->tv_hnsec      = (time0 >> 20) & 0xf;
   pTmS->tv_status     = (time0 >> 24) & 0xf;

}

/*****************************************************************************
* rdTime - read time from the board in days through seconds format. Read year
*/


void rdTime(int unit, uint pTmReg, vm_offset_t pY, stfp_time_t *pTmS)
{
   unsigned long time0, time1;
   int datain;
   long dpr_value, dpr_cmd;
   unsigned int days,icen,idec,ia,totaldays,iyr;

   struct stfp_softc *scp = stfp_softc[unit];


   STFP_RD_BUS_D8(scp->Dprhandle + pY, &dpr_value );
    pTmS->tm.tm_year = 0xff00 & (dpr_value << 8);
   STFP_RD_BUS_D8(scp->Dprhandle + pY +1, &dpr_value );
    pTmS->tm.tm_year |= 0xff & (dpr_value);

   STFP_RD_BUS_D32(scp->Regshandle + TIME0, &time0);
   STFP_RD_BUS_D32(scp->Regshandle + TIME1, &time1);

/*   printf("time0 = %08x time1 = %08x year = %d\n",time0,time1,pTmS->tm.tm_year);*/
  
   pTmS->tm.tm_yday =  time1 / 86400;

   pTmS->tm.tm_hour =  (time1 - pTmS->tm.tm_yday*86400)    / 3600;
 
   pTmS->tm.tm_min  =  (time1 - pTmS->tm.tm_yday*86400 - 
                                pTmS->tm.tm_hour*3600 )    / 60;

   pTmS->tm.tm_sec  =  (time1 - pTmS->tm.tm_yday*86400 - 
                                pTmS->tm.tm_hour*3600  - 
                                pTmS->tm.tm_min*60);
  
   pTmS->tm_usec       =  time0 & 0xfffff;
   pTmS->tm_hnsec      = (time0 >> 20) & 0xf;
   pTmS->tm_status     = (time0 >> 24) & 0xf;

/* Find the number of days from 1970 to the beginning of the current year */

totaldays = 0;
for(iyr=1970;iyr<pTmS->tm.tm_year;iyr++){
 days = 365;
 icen = iyr / 100;
 idec = iyr - 100 * icen;
 if(idec == 0)idec = icen;
 ia = idec / 4;
 if ((4 * ia) == idec)days=366;
 totaldays = totaldays+days;
}
/* Calc day of year */

pTmS->tm.tm_yday = pTmS->tm.tm_yday - totaldays;

}


/*****************************************************************************
* sendCommand - send command to the board to take action
*/
int sendCommand(int unit)
{
   struct stfp_softc *scp = stfp_softc[unit];
   int i;
   long dpr_value;

   STFP_REG_PUTL(ACK, ACK_COMMAND);	/* clear ack command done bit */
   STFP_REG_PUTL(ACK, ACK_ACTION);	/* force bc635PCI to take action */

   
   for (i=0;((STFPRegGetL(scp->Regshandle+ACK) & ACK_COMMAND) == 0) && i<20; i++)


   if(i >= 20)				/* check if timeout occured */
   {
      if (stfp_developer_debug)
         printf ("sendCommand() timeout\n");
      return (EIO);
   }

   return (0);
}


