
/* stfp_io.h - bc635S device driver header */
/* Version 1.0: 12 May 94 */
/* Version 1.1: 18 Jan 95 */
/* Version 1.2: 05 May 98 */
/* Version 1.3: 04 Aug 98 jmk Northrop Grumman Corp Hawthorne, CA */

#ifndef	_STFP_IO_H
#define _STFP_IO_H

#include <sys/time.h>

/*--------------------------- data structures ---------------------------*/

struct	stfp_tm 
{
   unsigned int tm_sec;
   unsigned int tm_min;
   unsigned int tm_hour;
   unsigned int tm_mday;
   unsigned int tm_mon;
   unsigned int tm_year;
   unsigned int tm_wday;
   unsigned int tm_yday;
   unsigned int tm_isdst;
};


typedef struct periodic 
{
   u_char  sync1pps;			/* TRUE = sync periodic to 1pps */
   u_short n1;				/* divider n1 */
   u_short n2;				/* divider n2 */
}stfp_periodic_t;
typedef stfp_periodic_t *PerPtr;

/* Additional Structures to interpret GETDATA command results */
/* GETDATA_UTCINFO	0x18 */

struct gd_utcinfo 
{
   u_char       enable;
   char         leapsec;
   char         leapevt;
   unsigned int evttime;
};

/*  GETDATA_FWVER	0x1F */

struct gd_fwversion 
{
   u_char   vmajor;			/* major version number */
   u_char   vminor;			/* minor version number */
   u_char   month;			/* month (1 - 12) */
   u_char   day;			/* day (1 - 31 ) */
   u_short  year;			/* year (e.g. 1994) */
};

/*	GETDATA_OSCCTL	0x23 */

struct gd_oscdisctl 
{
   char     discctl;
   char     phasectl;
   u_short  phasestep;
   short    gain;
   float     con;
};

/* Passed as arg in the GETDATA (38) command */

struct getdata_t 
{
   u_char arg; 				/* Get Data Command Argument */
   u_char padding[3];
   union
   {
      u_char tmode;                	/* GETDATA_MODE      0x10 */
      u_char tformat;              	/* GETDATA_TFORMAT   0x11 */
      u_short year;                	/* GETDATA_YEAR      0x13 */
      struct periodic period;      	/* GETDATA_PER       0x14 */
      char tcformat;               	/* GETDATA_TCFMT     0x15 */
      char tcmod;                  	/* GETDATA_TCMOD     0x16 */
      int offset;                  	/* GETDATA_TOFF      0x17 */ 
      struct gd_utcinfo utc;       	/* GETDATA_UTCINFO   0x18 */
      struct gd_fwversion version; 	/* GETDATA_FWVER     0x1F */
      char clksrc;                 	/* GETDATA_CLKSRC    0x20 */
      char jsctl;                  	/* GETDATA_JAMSC     0x21 */
      struct gd_oscdisctl ocsctl;  	/* GETDATA_OSCCTL    0x23 */
      u_short davalue;             	/* GETDATA_DAVAL     0x24 */
   } data;
};


typedef struct stfp_time 
{
   struct stfp_tm tm;
   unsigned int tm_usec;				/* usec 0 - 999999 */
   unsigned int tm_hnsec;			/* hnsec 0 - 9 (hundreds of nsecs) */
   unsigned int tm_status;
}stfp_time_t;
typedef stfp_time_t *TmPtr;


typedef struct stfp_timeval 
{
   struct stfp_tm tv;
   unsigned int tv_usec;				/* usec 0 - 999999 */
   unsigned int tv_hnsec;
   unsigned int tv_status;
}stfp_timeval_t;
typedef stfp_timeval_t *TvPtr;

typedef struct stfp_version 
{
   u_char   vmajor;			/* major version number */
   u_char   vminor;			/* minor version number */
   u_char   month;			/* month (1 - 12) */
   u_char   day;			/* day (1 - 31 ) */
   u_short  year;			/* year (e.g. 1994) */
}stfp_version_t;
typedef stfp_version_t *VerPtr;

#define TM_YEAR_BASE 	1900
#define STFP_MAX_READ	80
#define STFP_MAX_WRITE	80

/*-------------------------- ioctl() commands -------------------------*/
/* IOW user to kernel  IOR Kernel to user (status)   
   IOWR Both Ways  IO no data xfer                 */




#define TIMEREQUEST	 _IO(0,1)   /* Request time snapshot */

#define EVENTREQUEST	 _IO(0,2)   /* Request time snapshot of event */

#define RDTIME		 _IOR(0,3,stfp_time_t)   /* Read snapshot time  */

#define RDEVENT		 _IOR(0,4,stfp_time_t)   /* Read event snapshot time  */

#define RDTIMEREQ	 _IOR(0,5,stfp_time_t)   /* Request Snapshot and read time  */


#define CONTROLTIMEBASE	 _IOW(0,6,u_short)   /* Setup and control timebase parameters  */
#define CLOCK_INTERNAL		1
#define CLOCK_EXTERNAL		2
#define JAMSYNC_ENABLE		3
#define JAMSYNC_DISABLE		4
#define DISCIPLINE_ENABLE	5
#define DISCIPLINE_DISABLE	6
#define FORCE_JAMSYNC		7
#define CLEAR_USEC		8

#define SETDAC	 	 _IOW(0,7,int)   /* Set digital to analog converter  */
#define SETDISCGAIN	 _IOW(0,8,int)   /* Set diSCIPLING GAIN  */

#define SETTIMINGOFFSET	 _IOW(0,9,int)   /* Select Timing Offset  */


#define SELTIMINGMODE	 _IOW(0,10,unsigned char)   /* Select Timing Mode  */
#define MODE_TIMECODE		0
#define MODE_FREERUN		1
#define MODE_EXT1PPS		2
#define MODE_GPS		6

#define SELTIMEFORMAT	 _IOW(0,11,u_char)   /* Select Timing Format  */
#define TIME_DECIMAL		0
#define TIME_BINARY		1

#define SELTCFORMAT	 _IOW(0,12,u_char)   /* Select Timecode Format  */
#define TC_IRIGA		'A'
#define TC_IRIGB		'B'
#define TC_2137			'C'
#define TC_NASA36		'N'
#define TC_XR3			'X'

#define SELTCMOD	 _IOW(0,13,u_char)   /* Select IRIG type  */
#define MOD_AM			'M'
#define MOD_DCLS		'D'

#define SETPERIODIC	 _IOW(0,14,stfp_periodic_t)   /* Select SYNC type */
#define PERIODIC_NOSYNC		0
#define PERIODIC_SYNC		1

#define CONTROLEVENT	 _IOW(0,15,u_char)   /* Select Control Event */
#define EVENT_RISING		1
#define EVENT_FALLING		2
#define CAPLOCK_ENABLE		3
#define CAPLOCK_DISABLE		4
#define EVCAP_ENABLE		5
#define EVCAP_DISABLE		6
#define EVCAP_PERIODIC		7
#define EVCAP_EVENT		8
#define READ_CONTROL            9 
#define CLEAR_CONTROL           10
#define SET_CONTROL             11


#define CONTROLSTROBE	 _IOW(0,16,u_char)   /* Select Strobe Control */
#define STROBE_SECUS		1
#define STROBE_USONLY		2
#define STROBE_ENABLE		3
#define STROBE_DISABLE		4

#define SELFREQUENCYOUT	 _IOW(0,17,u_char)   /* Select Frequency Out */
#define FREQ_10MHZ		1
#define FREQ_5MHZ		2
#define FREQ_1MHZ		3

#define SIGNALCONTROL        _IOW(0,18,u_char)
#define DISABLE_PSIGNAL   0
#define ENABLE_PSIGNAL    1

#define GETDATA	 	 _IOWR(0,19,struct gd_oscdisctl)   /* data length equals longest structure Get requested values */
#define GETDATA_MODE		0x10
#define GETDATA_TFORMAT		0x11
#define GETDATA_YEAR		0x13
#define GETDATA_PER		0x14
#define GETDATA_TCFMT		0x15
#define GETDATA_TCMOD		0x16
#define GETDATA_TOFF		0x17
#define GETDATA_UTCINFO		0x18
#define GETDATA_FWVER		0x1F
#define GETDATA_CLKSRC		0x20
#define GETDATA_JAMSC		0x21
#define GETDATA_OSCCTL		0x23
#define GETDATA_DAVAL		0x24

#define RDEVENTREQ	 _IOR(0,20,stfp_timeval_t)   /* Request Event Snapshot and read event time  */


#define RDTIMETV	 _IOR(0,21,stfp_timeval_t)   

#define RDEVENTTV	 _IOR(0,22,stfp_timeval_t)   

#define RDTIMETVREQ	 _IOR(0,23,stfp_timeval_t)   

#define RDEVENTTVREQ	 _IOR(0,24,stfp_timeval_t)   


#define CAPUNLOCK	 _IO(0,25)   /*  */

#define RDINTSTAT	 _IOR(0,26,unsigned int)   /*  */

#define CLRINTSTAT	 _IOW(0,27,unsigned int)   /*  */

#define WRSTROBE	 _IOW(0,28,stfp_time_t)   /*  */

#define WRSTROBETV	 _IOW(0,29,stfp_timeval_t)   /*  */

#define SETTIME	 	_IOW(0,30,stfp_time_t)   /*  */

#define SETTIMETV	 _IOW(0,31,stfp_timeval_t)  

#define SETYEAR	 	_IOW(0,32,unsigned int)   /*   */

#define SETINTSIGNAL	 _IOW(0,33,unsigned int)   /*  */

#define RDINTSIGNAL	 _IOR(0,34,unsigned int)   /*  */

#define RDINTTIME	 _IOR(0,35,stfp_timeval_t)   /*  */

#define RESETTFP         _IO(0,36)

#define SETRTSIGNAL      _IOW(0,37,int)


#define COMMANDGETDATA	 _IOWR(0,38,struct gd_oscdisctl)   /* data length equals longest structure Get requested values */
#define COM_GETDATA       0x19

#define SETLOCALOFFSET		20




#define SELFCONN		24

#define FCONN_EVENT_IN		0x00
#define FCONN_1PPS_IN		0x01
#define FCONN_DCLS_IN		0x02
#define FCONN_STR2EV		0x03
#define FCONN_1PPS_OUT		0x08
#define FCONN_FREQ_OUT		0x09
#define FCONN_PER_OUT		0x0A
#define FCONN_STROBE_OUT	0x0B
#define FCONN_DCLS_OUT		0x0C


/*------------------------- bc635S DPRAM Commands -------------------------*/

#define COM_RESET               0x1A
#define COM_SETTIMINGMODE	0x10

#define COM_SELTIMEFORMAT	0x11
#define COM_SETTIME		0x12
#define COM_SETYEAR		0x13
#define COM_SETPERIODIC		0x14
#define COM_SETTCFORMAT		0x15
#define COM_SETTCMOD		0x16
#define COM_SETTIMINGOFFSET	0x17


#define COM_SELCLOCK		0x20
#define SELCLOCK_INT		'I'
#define SELCLOCK_EXT		'E'

#define COM_JAMSYNCENABLE	0x21
#define JAMSYNC_DIS		0
#define JAMSYNC_EN		1
#define COM_FORCEJAMSYNC	0x22

#define	COM_DISCENABLE		0x23
#define DISC_DIS		0
#define DISC_EN			1

#define COM_SETDAC		0x24
#define COM_SETDISCGAIN		0x25
#define COM_SENDGPSPACKET	0x30
#define COM_REQGPSPACKET	0x31
#define COM_MANREQGPSPACKET	0x32
#define COM_GPSTIMEFORMAT	0x33

#endif	/* _STFP_IO_H */

