
/*-----------------------------------------------------------------------
 * stfp_reg.h - bc635S device driver header 
 * 
 * Version 1.0  created 05 Aug 98 jmk Northrop Grumman Corp Hawthorne, CA
 *-----------------------------------------------------------------------
 */

/*-----------------------------------------------------------------------
 * Tables and Section Numbers referenced in this file are from the DATUM, 
 * INC "bc635PCI TIME & FREQUENCY PROCESSOR User's Guide" - December 1997
 *-----------------------------------------------------------------------
 */

#ifndef _STFP_REG_H_
#define _STFP_REG_H_

#ident "@(#)stfp_reg.h 1.0 98/08/04 DATUM"

/*
 *------------------------------------------------------------------------
 * The DATUM_ID symbol defines the PCI vendor ID code assigned by Digital
 * Equipment Corporation. Don't change it!
 *------------------------------------------------------------------------
 */

#define DATUM_ID        0x12e2  /* Datum Inc., Bancomm-Timing Division */

/*
 *------------------------------------------------------------------------
 * Data elements for portability
 *------------------------------------------------------------------------
 */

typedef unsigned char		Byte;	/* 8-bits wide			 */
typedef	unsigned short int	Word;	/* 16-bits wide			 */
typedef unsigned int		Lword;	/* 32-bits wide			 */

/*-------------------------------------------------------------------------
 * CSR I/O access macros for ioctl data transfers
 * Uses the _IOR macro to construct ioctl macros:
 *
 * DS_GETCOUNT defines ioctl types for situations where data is transferred
 * from the kernel to the user's buffer
 *
 * DS_CLRCOUNT defines ioctl types for situations where no data is actually
 * transferred from the kernel to the user's buffer
 *-------------------------------------------------------------------------
 */

#define DS_GETCOUNT	_IOR(0,1,int)	
#define DS_CLRCOUNT	_IO(0,2)

#define STFP_CSR	0

#define TOP_DPRAM 0x800         /* top of Dual Port Ram */

#define romsize  0x800          /* Dual Port Ram Starts at top of ROM */




/*-----------------------------------------------------------------------
 * bc635S STATUS bits (from TIME0, EVENT0 registers) Table 5-7
 *-----------------------------------------------------------------------
 */

#define STAT_FLYWHEELING	0x01
#define STAT_TIME_OFFSET	0x02
#define STAT_FREQ_OFFSET	0x04

/*-----------------------------------------------------------------------
 * bc635S CONTROL register bits Table 5-3
 *-----------------------------------------------------------------------
 */ 

#define CTRL_LOCKEN		0x01
#define CTRL_EVSOURCE		0x02
#define CTRL_EVSENSE		0x04
#define CTRL_EVENTEN		0x08
#define CTRL_STREN		0x10
#define CTRL_STRMODE		0x20
#define CTRL_F10		0x00
#define CTRL_F5			0x40
#define CTRL_F1			0x80
#define CTRL_FMASK		0xC0

/*-----------------------------------------------------------------------
 * bc635S ACK register bits Section 6.1
 *-----------------------------------------------------------------------
 */

#define ACK_COMMAND		0x01
#define ACK_PACKET		0x04
#define ACK_ACTION		0x80

/*-----------------------------------------------------------------------
 * bc635S INTSTAT and MASK register bits  Table 5-4
 *-----------------------------------------------------------------------
 */ 
		
#define INT_EVENT		0x01
#define INT_PERIODIC		0x02
#define INT_STROBE		0x04
#define INT_1PPS		0x08
#define INT_PACKET		0x10
#define INT_BITMASK		0x1F
#define INT_CLEAR               0x00

/*-----------------------------------------------------------------------
 * bc635 TFP DEVICE REGISTER Table 5-2
 * The following areas are Reserved:  0x0C, 0x28 & 0x2C
 *-----------------------------------------------------------------------
 */
 
#define TIMEREQ			0x00	/* Time Request (TIME0-7)	*/
#define EVENTREQ		0x04	/* Event Request (EVENT0-4)	*/
#define RELEASE_CAP		0x08	/* Release Capture Lockout	*/
#define CONTROL_REG		0x10	/* Control Register		*/
#define ACK			0x14	/* Acknowledge Register		*/
#define MASK			0x18	/* Interrupt Mask		*/
#define INTSTAT			0x1C	/* Interrupt Status		*/
#define MINSTRB			0x20	/* Minor strobe Time		*/
#define MAJSTRB			0x24	/* Major Strobe Time		*/
#define TIME0			0x30	/* Minor Time Holding Register	*/
#define TIME1			0x34	/* Major Time Holding Register	*/
#define EVENT0			0x38	/* Minor Event Holding Register	*/
#define EVENT1			0x3C	/* Major Event Holding Register	*/

#endif /* _STFP_REG_H_ */


